//
//  NSDirectoryServices.m
//

/*
 * Extreme Mac OS X Programming
 *
 * Using Google Web APIs in a Mac OS X Application
 * http://javworld.com/extreme/GoogleAPI.html
 *
 * Copyright © 2004 John A. Vink
 *
 */

#import "NSDirectoryServices.h"

#import <DirectoryService/DirServices.h>
#import <DirectoryService/DirServicesUtils.h>
#import <DirectoryService/DirServicesConst.h>


@interface NSDirectoryServices (Private)
- (BOOL) openNode:(NSString*)node;
- (void) addRecordAtIndex:(unsigned long)recIndex toDictionary:(NSMutableDictionary*)dict withBuffer:(tDataBufferPtr)dataBuffer;
- (void) addEntry:(tAttributeEntry*)pAttrEntry withValue:(tAttributeValueListRef)valueRef toDictionary:(NSMutableDictionary*)dict withBuffer:(tDataBufferPtr)dataBuffer;
- (NSString*) nameOfRecord:(tRecordEntry*)recEntry attributes:(tAttributeListRef)attrListRef withBuffer:(tDataBufferPtr)dataBuffer;
@end


@implementation NSDirectoryServices

// ---------------------------------------------------------------------
// * init
// ---------------------------------------------------------------------
- (id) init
{
	if (nil != (self = [super init]))
	{
			// init directory services
		if (eDSNoErr != dsOpenDirService(&_dirRef))
		{
			printf("Can't open directory services\n");
			return nil;
		}
	}
	
	return self;
}


// ---------------------------------------------------------------------
// * dealloc
// ---------------------------------------------------------------------
- (void) dealloc
{
		// close directory services
	dsCloseDirService(_dirRef);
	
		// this ain't C++.  Always call your parent's dealloc, or you're not releasing any memory.
		// Only losers don't dealloc memory.  Don't be a loser.
	[super dealloc];
}


// ---------------------------------------------------------------------
// * getNodes
// ---------------------------------------------------------------------
// this returns a list of the available nodes.
// usually you'll just want to use the first one.
- (NSArray*) getNodes
{
    bool done = false;
    tDirStatus dirStatus = eDSNoErr;
    unsigned long bufferCount = 0;
    tDataBufferPtr dataBuffer = nil;
    tDataListPtr nodeName = nil;
    tContextData context = nil;
	NSMutableArray* result = [NSMutableArray array];

		// allocate a 32k buffer.
	dataBuffer = dsDataBufferAllocate(_dirRef, 32 * 1024);
	if (dataBuffer != nil)
	{
		while ((dirStatus == eDSNoErr) && !done)
		{
			dirStatus = dsFindDirNodes( _dirRef, dataBuffer, nil, eDSAuthenticationSearchNodeName, 
					&bufferCount, &context );
			if (dirStatus == eDSNoErr && bufferCount > 0)
			{
				dirStatus = dsGetDirNodeName( _dirRef, dataBuffer, 1, &nodeName );
				if (dirStatus == eDSNoErr)
				{
					char* pNodeName = dsGetPathFromList(_dirRef, nodeName, "/");
					[result addObject:[NSString stringWithCString:pNodeName]];
					dsDataListDeallocate(_dirRef, nodeName);
					free(nodeName);
				}
				else
				{
					NSLog(@"dsGetDirNodeName error = %ld, count = %ld", dirStatus, bufferCount );
				}
			}
			else
				NSLog(@"Error %ld from dsFindDirNodes", dirStatus);
			done = (context == nil);
		}
		dsDataBufferDeAllocate( _dirRef, dataBuffer );
	}
	
		// remember this list of nodes for future function calls
	_nodes = [[NSArray arrayWithArray:result] retain];
	
		// return a copy to the caller
	return [_nodes copy];
}


// ---------------------------------------------------------------------
// * getRecord
// ---------------------------------------------------------------------
- (NSDictionary*) getRecord:(NSString*)recordName andType:(NSString*)recordType fromNode:(NSString*)node
{
	tDirStatus				dirStatus = eDSNoErr;
	NSMutableDictionary*	result = [NSMutableDictionary dictionary];
	
		// if the user did not give us a node, try and get the first one
	if (node == nil)
	{
			// see if any nodes were ever fetched
		if (_nodes == nil)
			[self getNodes];
			// see if there are any nodes
		if (_nodes && [_nodes count] == 0)
		{
			NSLog(@"No nodes available");
			return nil;
		}
		
			// use the first node we find
		node = [_nodes objectAtIndex:0];
	}

	do
	{
		if (![self openNode:node])
			break;
			
		tDataList recNames = {};
		tDataList recTypes = {};
		tDataList attrTypes = {};

		if ( (eDSNoErr == (dirStatus = dsBuildListFromStringsAlloc (_dirRef, &recNames, [recordName cString], nil))) &&
			 (eDSNoErr == (dirStatus = dsBuildListFromStringsAlloc (_dirRef, &recTypes, [recordType cString], nil))))
		{
			dirStatus = dsBuildListFromStringsAlloc (_dirRef, &attrTypes, kDSAttributesAll, nil);
		}
		
		if (dirStatus != eDSNoErr)
			break;
		
		tDataBufferPtr dataBuffer = dsDataBufferAllocate(_dirRef, 32 * 1024);
		if (dataBuffer == nil)
			break;
			
		tContextData context = nil;
		unsigned long recCount;
		dirStatus = dsGetRecordList(_nodeRef, dataBuffer, &recNames, eDSExact, &recTypes, &attrTypes, false, &recCount, &context);
		if (dirStatus != eDSNoErr)
		{
			NSLog(@"Can't get record list");
			break;
		}
			
		if (recCount < 1)
		{
			NSLog(@"Number records: %ld", recCount);
			break;
		}
		
		[self addRecordAtIndex:1 toDictionary:result withBuffer:dataBuffer];

		// Deallocate recNames, recTypes, and attrTypes by calling dsDataListDeallocate.
		dsDataListDeallocate(_dirRef, &recNames);
		dsDataListDeallocate(_dirRef, &recTypes);
		dsDataListDeallocate(_dirRef, &attrTypes);
		dsDataBufferDeAllocate(_dirRef, dataBuffer);

	} while (0);
	
	return [NSDictionary dictionaryWithDictionary:result];
}


// ---------------------------------------------------------------------
// * getRecord
// ---------------------------------------------------------------------
- (NSArray*) getRecordsOfType:(NSString*)recordType fromNode:(NSString*)node
{
	tDirStatus				dirStatus = eDSNoErr;
	NSMutableArray*			result = [NSMutableArray new];
	
		// if the user did not give us a node, try and get the first one
	if (node == nil)
	{
			// see if any nodes were ever fetched
		if (_nodes == nil)
			[self getNodes];
			// see if there are any nodes
		if (_nodes && [_nodes count] == 0)
		{
			NSLog(@"No nodes available");
			return nil;
		}
		
			// use the first node we find
		node = [_nodes objectAtIndex:0];
	}

	do
	{
		if (![self openNode:node])
			break;
			
		tDataList recNames = {};
		tDataList recTypes = {};
		tDataList attrTypes = {};

		if ((eDSNoErr == (dirStatus = dsBuildListFromStringsAlloc (_dirRef, &recNames, kDSRecordsAll, nil))) &&
			(eDSNoErr == (dirStatus = dsBuildListFromStringsAlloc (_dirRef, &recTypes, [recordType cString], nil))))
		{
			dirStatus = dsBuildListFromStringsAlloc (_dirRef, &attrTypes, kDSAttributesAll, nil);
		}
		
		if (dirStatus != eDSNoErr)
			break;
		
		tDataBufferPtr dataBuffer = dsDataBufferAllocate(_dirRef, 32 * 1024);
		if (dataBuffer == nil)
			break;
			
		tContextData context = nil;
		unsigned long recCount;
		dirStatus = dsGetRecordList(_nodeRef, dataBuffer, &recNames, eDSExact, &recTypes, &attrTypes, false, &recCount, &context);
		if (dirStatus != eDSNoErr)
		{
			NSLog(@"Can't get record list.  Error %ld", dirStatus);
			break;
		}
			
		if (recCount < 1)
		{
			NSLog(@"Number records: %ld", recCount);
			break;
		}
		
		unsigned long i;
		tRecordEntry* recEntry = nil;
		tAttributeListRef attrListRef = nil;
		for (i = 1; i <= recCount; i++)
		{
			dirStatus = dsGetRecordEntry(_nodeRef, dataBuffer, i, &attrListRef, &recEntry);
			NSString* recordName = [self nameOfRecord:recEntry attributes:attrListRef withBuffer:dataBuffer];
			if (recordName)
				[result addObject:recordName];
		}

		// Deallocate recNames, recTypes, and attrTypes by calling dsDataListDeallocate.
		dsDataListDeallocate(_dirRef, &recNames);
		dsDataListDeallocate(_dirRef, &recTypes);
		dsDataListDeallocate(_dirRef, &attrTypes);
		dsDataBufferDeAllocate(_dirRef, dataBuffer);

	} while (0);
	
	return [NSArray arrayWithArray:result];
}


@end


@implementation NSDirectoryServices (Private)

// ---------------------------------------------------------------------
// * openNode
// ---------------------------------------------------------------------
- (BOOL) openNode:(NSString*)node
{
    tDataListPtr			nodePath = nil;
	tDirStatus				dirStatus;

	nodePath = dsBuildFromPath(_dirRef, [node cString], "/" );
	if (nodePath == nil)
		return NO;
		
	BOOL result = YES;
	dirStatus = dsOpenDirNode(_dirRef, nodePath, &_nodeRef);
	if (dirStatus != eDSNoErr)
	{
		NSLog(@"Can't open node");
		result = NO;
	}
	dsDataListDeallocate(_dirRef, nodePath);
	free(nodePath);
	
	return result;
}


// ---------------------------------------------------------------------
// * addRecordAtIndex
// ---------------------------------------------------------------------
// add the record to the dictionary
- (void) addRecordAtIndex:(unsigned long)recIndex toDictionary:(NSMutableDictionary*)dict withBuffer:(tDataBufferPtr)dataBuffer
{
	tDirStatus dirStatus;
	short i;

	tRecordEntry* recEntry = nil;
	tAttributeListRef attrListRef = nil;
	dirStatus = dsGetRecordEntry(_nodeRef, dataBuffer, 1, &attrListRef, &recEntry);
	if (dirStatus == eDSNoErr)
	{
		for (i = 1; i <= recEntry->fRecordAttributeCount; i++)
		{
			tAttributeValueListRef valueRef = nil;
			tAttributeEntry *pAttrEntry = nil;
			if (eDSNoErr == (dirStatus = dsGetAttributeEntry(_nodeRef, dataBuffer, attrListRef, i, &valueRef, &pAttrEntry)))
			{
				[self addEntry:pAttrEntry withValue:valueRef toDictionary:dict withBuffer:(tDataBufferPtr)dataBuffer];
				
				dirStatus = dsDeallocAttributeEntry(_dirRef, pAttrEntry);
			}
		}
		dsCloseAttributeList(attrListRef);
		dsDeallocRecordEntry(_dirRef, recEntry);
	}
	else
		NSLog(@"Can't get record entry - %ld", dirStatus);
}


// ---------------------------------------------------------------------
// * addEntry
// ---------------------------------------------------------------------
// Adds an entry to the dictionary.  The entry could be a single value or an array of values.
- (void) addEntry:(tAttributeEntry*)pAttrEntry withValue:(tAttributeValueListRef)valueRef toDictionary:(NSMutableDictionary*)dict withBuffer:(tDataBufferPtr)dataBuffer
{
	tDirStatus dirStatus;
	tAttributeValueEntry *pValueEntry = nil;

		// if there is only one value, then we can add it to the result as a string
	if (pAttrEntry->fAttributeValueCount == 1)
	{
		if (eDSNoErr == (dirStatus = dsGetAttributeValue(_nodeRef, dataBuffer, 1, valueRef, &pValueEntry)))
		{
			[dict setObject:[NSString stringWithCString:pValueEntry->fAttributeValueData.fBufferData] forKey:[NSString stringWithCString:pAttrEntry->fAttributeSignature.fBufferData]];						
			dirStatus = dsDeallocAttributeValueEntry(_dirRef, pValueEntry);
		}
	}
		// if there is more than one value, then we need to add it to the result as an array
	else if (pAttrEntry->fAttributeValueCount > 1)
	{
		NSMutableArray* entries = [NSMutableArray new];
		unsigned long k;
			// iterate over all the entries and add them to the array
		for (k = 1; k <= pAttrEntry->fAttributeValueCount; k++)
		{
			if (eDSNoErr == (dirStatus = dsGetAttributeValue(_nodeRef, dataBuffer, k, valueRef, &pValueEntry)))
			{
				[entries addObject:[NSString stringWithCString:pValueEntry->fAttributeValueData.fBufferData]];							
				dirStatus = dsDeallocAttributeValueEntry(_dirRef, pValueEntry);
			}
		}
		
		[dict setObject:[NSArray arrayWithArray:entries] forKey:[NSString stringWithCString:pAttrEntry->fAttributeSignature.fBufferData]];
		[entries release];
	}
}


// ---------------------------------------------------------------------
// * nameOfRecord
// ---------------------------------------------------------------------
// looks through all the attributes and finds the name attribute.  And returns it.
- (NSString*) nameOfRecord:(tRecordEntry*)recEntry attributes:(tAttributeListRef)attrListRef withBuffer:(tDataBufferPtr)dataBuffer
{
	tDirStatus dirStatus = eDSNoErr;
	unsigned long i;

		// iterate through all the attributes looking for the name attribute
	for (i = 1; i <= recEntry->fRecordAttributeCount; i++)
	{
		tAttributeValueListRef valueRef = nil;
		tAttributeEntry *pAttrEntry = nil;
		
			// get the attribute
		if (eDSNoErr == (dirStatus = dsGetAttributeEntry(_nodeRef, dataBuffer, attrListRef, i, &valueRef, &pAttrEntry)))
		{
				// cehck to see if this is the name attribute
			if (0 == strcmp(pAttrEntry->fAttributeSignature.fBufferData, kDSNAttrRecordName))
			{
					// this is the name attribute.  Now fetch the value and return it.
				tAttributeValueEntry *pValueEntry = nil;
				NSString* result = nil;
				if (eDSNoErr == (dirStatus = dsGetAttributeValue(_nodeRef, dataBuffer, 1, valueRef, &pValueEntry)))
					result = [NSString stringWithCString:pValueEntry->fAttributeValueData.fBufferData];

				dsDeallocAttributeEntry(_dirRef, pAttrEntry);
				return result;
			}
								
			dsDeallocAttributeEntry(_dirRef, pAttrEntry);
		}
	}

		// no name was found.  This should never happen.
	return nil;
}

@end
